﻿using System.Collections.Generic;
using System.Collections.Specialized;
using System.Linq;
using System.Reflection;
using System.Windows;
using Infragistics.Controls.Editors;

namespace WpfApp2
{
    public partial class MainWindow
    {
        private const string SourceObjectPropName = "SourceObject";
        private bool isAutoExpanding = false;
        private List<string> trackingPath = new List<string>();
        public MainWindow()
        {
            InitializeComponent();
            PropertyGrid.SelectedItemChanged += OnSelectedItemChanged;
            (PropertyGrid.Items as INotifyCollectionChanged).CollectionChanged += OnItemsChanged;
            //PropertyGrid.SelectedObject = new Sub();
            //WinPropertyGrid.SelectedObject = new Sub();
            PropertyGrid.SelectedObject = GenerateData();
        }

        private IList<IDetailsViewModel> GenerateData()
        {
            var result = new List<IDetailsViewModel>();
            var details = new DetailsViewModel
            {
                Name = "Some category"
            };

            details.Properties.Add(new PropertyViewModel
            {
                Name = "simple1",
                Value = 2,
                ValueType = typeof(int)
            });

            details.Properties.Add(new PropertyViewModel
            {
                Name = "simple2",
                Value = "some text",
                ValueType = typeof(string)
            });

            details.Properties.Add(new PropertyViewModel
            {
                Name = "expandable ok",
                Properties = new List<IPropertyViewModel>
                {
                    new PropertyViewModel
                    {
                        Name = "expandable further not ok",
                        Properties = new List<IPropertyViewModel>
                        {
                            new PropertyViewModel
                            {
                                Name = "some prop",
                                ValueType = typeof(string),
                                Value = "some value"
                            }
                        }
                    }
                }
            });

            details.Properties.Add(new PropertyViewModel
            {
                Name = "expandable not ok",
                Properties = new List<IPropertyViewModel>
                {
                    new PropertyViewModel
                    {
                        Name = "some prop",
                        ValueType = typeof(string),
                        Value = "some value"
                    },
                    new PropertyViewModel
                    {
                        Name = "expandable further",
                        Properties = new List<IPropertyViewModel>
                        {
                            new PropertyViewModel
                            {
                                Name = "some prop",
                                ValueType = typeof(string),
                                Value = "some value"
                            }
                        }
                    }
                }
            });

            result.Add(details);

            return result;
        }

        private void OnItemsChanged(object sender, NotifyCollectionChangedEventArgs e)
        {
            ExpandPath();
        }

        private void OnSelectedItemChanged(object sender, SelectedItemChangedEventArgs e)
        {
            if (e.OldSelectedItem is PropertyGridPropertyItem item)
            {
                SetDisplayName(item);
            }

            if (e.NewSelectedItem is PropertyGridPropertyItem item2)
            {
                SetDisplayName(item2);
                if (isAutoExpanding)
                {
                    return;
                }

                PropertyGridItem pathItem = item2;
                trackingPath.Clear();
                while (pathItem != null)
                {
                    trackingPath.Add(pathItem.GetPropertyValue<IPropertyViewModel>(SourceObjectPropName)?.Name ?? pathItem.DisplayName);
                    pathItem = pathItem.GetPropertyValue<PropertyGridItem>("ParentItem");
                }

                trackingPath.Reverse();
            }
        }

        private void ExpandPath()
        {
            if (trackingPath.Count == 0 || PropertyGrid.Items.Count == 0)
            {
                return;
            }

            isAutoExpanding = true;
            var itemToSelect = PropertyGrid.Items.FirstOrDefault(_ => _.DisplayName == trackingPath[0]);

            for (var i = 1; i < trackingPath.Count && itemToSelect != null; i++)
            {
                itemToSelect = itemToSelect.Children.FirstOrDefault(_ => _.DisplayName == trackingPath[i]);

                if (itemToSelect != null && i < trackingPath.Count - 1)
                {
                    itemToSelect.IsExpanded = true; // crashes app here
                }
            }

            if (itemToSelect != null)
            {
                PropertyGrid.SelectedItem = itemToSelect; //and/or crashes here
            }

            isAutoExpanding = false;
        }

        private static void SetDisplayName(PropertyGridPropertyItem item)
        {
            var prop = item.GetPropertyValue<IPropertyViewModel>("SourceObject");
            item.DisplayName = prop.Name;
        }

        private void ButtonBase_OnClick(object sender, RoutedEventArgs e)
        {
            PropertyGrid.SelectedObject = GenerateData();
        }
    }
    public static class ReflectionUtils
    {
        public static TValue GetPropertyValue<TValue>(this object instance, string propName) where TValue : class
        {
            return instance.GetType().GetProperty(propName, BindingFlags.Instance | BindingFlags.NonPublic)
                ?.GetValue(instance) as TValue;
        }
    }

    public class Base
    {
        public int Type { get; set; } = 4;
    }

    public class Sub : Base
    {
        private int x = 0;
        public new string Type { get; set; } = "asfdsafa af af ";
    }
}
